#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <string.h>
#include <stdlib.h>
#include "lcd.h"

#define swap(A, B) A = A + B; B = A - B; A = A - B;

#define LCD_WIDTH  84
#define LCD_HEIGHT 48

#define BUZZER PC0
#define ENTER_BTN PD0
#define RIGHT_BTN PD1
#define UP_BTN PD6
#define LEFT_BTN PD5
#define DOWN_BTN PD4

#define N 4
#define M 8
#define ROUNDS 3

#define X_START_GRID 2
#define Y_START_GRID 2
#define CELL_WIDTH 10
#define PADDING 2

unsigned char form[4][CELL_WIDTH - 1][CELL_WIDTH - 1] = {
	/* square */
   {{0, 0, 0, 0, 0, 0, 0, 0, 0},
	{0, 0, 0, 0, 0, 0, 0, 0, 0},
	{0, 0, 1, 1, 1, 1, 1, 0, 0},
	{0, 0, 1, 0, 0, 0, 1, 0, 0},
	{0, 0, 1, 0, 0, 0, 1, 0, 0},
	{0, 0, 1, 0, 0, 0, 1, 0, 0},
	{0, 0, 1, 1, 1, 1, 1, 0, 0},
	{0, 0, 0, 0, 0, 0, 0, 0, 0},
	{0, 0, 0, 0, 0, 0, 0, 0, 0}},
	
	/* arrow */
   {{0, 0, 0, 0, 0, 0, 0, 0, 0},
	{0, 0, 0, 0, 0, 0, 0, 0, 0},
	{0, 0, 0, 0, 0, 1, 0, 0, 0},
	{0, 0, 1, 0, 0, 1, 1, 0, 0},
	{0, 0, 1, 1, 1, 1, 1, 1, 0},
	{0, 0, 1, 0, 0, 1, 1, 0, 0},
	{0, 0, 0, 0, 0, 1, 0, 0, 0},
	{0, 0, 0, 0, 0, 0, 0, 0, 0},
	{0, 0, 0, 0, 0, 0, 0, 0, 0}},
	
	/* house */
   {{0, 0, 0, 0, 1, 0, 0, 0, 0},
	{0, 0, 0, 1, 0, 1, 0, 0, 0},
	{0, 0, 1, 0, 0, 0, 1, 0, 0},
	{0, 0, 1, 1, 1, 1, 1, 0, 0},
	{0, 0, 1, 0, 0, 0, 1, 0, 0},
	{0, 0, 1, 0, 0, 0, 1, 0, 0},
	{0, 0, 1, 0, 1, 0, 1, 0, 0},
	{0, 0, 1, 1, 1, 1, 1, 0, 0},
	{0, 0, 0, 0, 0, 0, 0, 0, 0}},
	
	/* circle */
   {{0, 0, 0, 0, 0, 0, 0, 0, 0},
	{0, 0, 0, 1, 1, 1, 0, 0, 0},
	{0, 0, 1, 1, 0, 1, 1, 0, 0},
	{0, 1, 1, 0, 0, 0, 1, 1, 0},
	{0, 1, 0, 0, 0, 0, 0, 1, 0},
	{0, 1, 1, 0, 0, 0, 1, 1, 0},
	{0, 0, 1, 1, 0, 1, 1, 0, 0},
	{0, 0, 0, 1, 1, 1, 0, 0, 0},
	{0, 0, 0, 0, 0, 0, 0, 0, 0}}};
unsigned char table[N][M];
unsigned char curX = 0, curY = 0;
unsigned char sel_cell = 0;
unsigned char game_over = 0;
unsigned char points = 0;
char message[30];


void restart_game()
{
	game_over = 0;
	points = 0;
	curX = 0;
	curY = 0;
	sel_cell = 0;
	
}

void btn_pressed()
{
	/* check if one of the directions is pressed
	   and if it pressed:
		1. update position of current cell
		2. check if we selected to move a cell, in which
		   case we will swap the figures
	*/
	if (!(PIND & (1 << UP_BTN)) && curX > 0) {
		curX--;
		if (sel_cell == 1) {
			swap(table[curX][curY], table[curX + 1][curY]);
			sel_cell = 0;
		}
		_delay_ms(150);
	}
	if (!(PIND & (1 << RIGHT_BTN)) && curY < M - 1) {
		curY++;
		if (sel_cell == 1) {
			swap(table[curX][curY], table[curX][curY - 1]);
			sel_cell = 0;
		}
		_delay_ms(150);
	}
	if (!(PIND & (1 << DOWN_BTN)) && curX < N - 1) {
		curX++;
		if (sel_cell == 1) {
			swap(table[curX][curY], table[curX - 1][curY]);
			sel_cell = 0;
		}
		_delay_ms(150);
	}
	if (!(PIND & (1 << LEFT_BTN)) && curY > 0) {
		curY--;
		if (sel_cell == 1) {
			swap(table[curX][curY], table[curX][curY + 1]);
			sel_cell = 0;
		}
		_delay_ms(150);
	}
	
	/* Enter is pressed */
	if (!(PIND & (1 << ENTER_BTN))) {
		
		/* check if we have a selected cell in queue */
		if (sel_cell == 0 && game_over < ROUNDS)
			sel_cell = 1;
		
		if (game_over >= ROUNDS) {
			restart_game();
		}
		_delay_ms(150);
	}
}

ISR(PCINT3_vect)
{
	btn_pressed();
}

ISR(TIMER1_COMPA_vect)
{
	game_over++;
}

ISR(TIMER1_COMPB_vect)
{
	
}

void init_btns() 
{	
	sei();
	/* ENTER */
	DDRD &= ~(1 << ENTER_BTN);
	PORTD |= (1 << ENTER_BTN);
	
	/* LEFT */
	DDRD &= ~(1 << LEFT_BTN);
	PORTD |= (1 << LEFT_BTN);
	
	/* UP */
	DDRD &= ~(1 << UP_BTN);
	PORTD |= (1 << UP_BTN);
	
	/* RIGHT */
	DDRD &= ~(1 << RIGHT_BTN);
	PORTD |= (1 << RIGHT_BTN);
	
	/* DOWN */
	DDRD &= ~(1 << DOWN_BTN);
	PORTD |= (1 << DOWN_BTN);
	
	// Activez PIN CHANGE INTERRUPT ENABLE pentru PCINT24:31 (portul D)
    PCICR |= (1 << PCIE3);
    // Activez intreruperea pe butoanele PB2 si PD6
    PCMSK3 |= (1 << PCINT24) | (1 << PCINT25) | (1 << PCINT28) | (1 << PCINT29) | (1 << PCINT30);
}
 
 void display_character(int x, int y, unsigned char type, int selected)
 {
	int px = X_START_GRID + x * CELL_WIDTH + 1;
	int py = Y_START_GRID + y * CELL_WIDTH + 1;
	size_t i, j;
	
	for (i = 0; i < CELL_WIDTH; i++)
		for (j = 0; j < CELL_WIDTH; j++)
			if (selected == 0 && form[type][i][j]  == 1)
				lcd_draw_pixel(py + j, px + i);
			else if (selected == 1 && form[type][i][j]  == 0)
				lcd_draw_pixel(py + j, px + i);
 }
 
 void display_line(int x, int y, int orientation)
 {
	size_t i;
	for (i = 0; i < CELL_WIDTH; i++) {
		lcd_draw_pixel(y, x);
		orientation == 0 ? y++ : x++;
	}
 }
 
 void display_grid()
 {
	size_t i, j;
	for (i = 0; i <= N; i++)
		for (j = 0; j < M; j++)
			display_line(X_START_GRID + i * CELL_WIDTH, Y_START_GRID + j * CELL_WIDTH, 0);
	for (i = 0; i < N; i++)
		for (j = 0; j <= M; j++)
			display_line(X_START_GRID + i * CELL_WIDTH, Y_START_GRID + j * CELL_WIDTH, 1);
 }

void init_table()
 {
	size_t i, j;
	for (i = 0; i < N; i++)
		for (j = 0; j < M; j++)
			table[i][j] = rand() % 4;
 }
 
void display_table()
{
	size_t i, j;
	for (i = 0; i < N; i++)
		for (j = 0; j < M; j++)
			display_character(i, j, table[i][j], i==curX && j==curY);
}

void buzz()
{
		/* BUZZER */
	DDRC |= (1 << BUZZER);

	PINC &= ~(1 << BUZZER);
	_delay_ms(40);
	PINC |= (1 << BUZZER);
	_delay_ms(20);

	DDRC &= ~(1 << BUZZER);
}

void check_allignment()
{
	int ok;
	size_t i, j;
	
	/* Line */
	for (i = 0; i < N; i++) {
		ok = 1;
		for (j = 0; j < M - 1; j++)
			if (table[i][j] != table[i][j+1]) {
				ok = 0;
				break;
			}
		if (ok == 1) {
			/* Line i must be destroyed */
			buzz();
			
			/* Replace the forms */
			for (j = 0; j < M; j++)
				table[i][j] = rand() % 4;
			
			/* Update points */
			points += 3;
		}
	}
	
	/* Column */
	for (i = 0; i < M; i++) {
		ok = 1;
		for (j = 0; j < N - 1; j++)
			if (table[j][i] != table[j + 1][i]) {
				ok = 0;
				break;
			}
		if (ok == 1) {
			/* Column i must be destroyed */
			buzz();
			
			/* Replace the forms */
			for (j = 0; j < N; j++)
				table[j][i] = rand() % 4;
			
			/* Update points */
			points += 1;
		}
	}
			
}

void init_timer()
{
    ICR1 = 65535;
    OCR1A = 40000;
    OCR1B = 0;

    TIMSK1 = (1 << OCIE1A) | (1 << OCIE1B);

    TCCR1B = (5 << CS10) | (1 << WGM12) | (1 << WGM13);
}

void display_game_over()
{
	itoa(points, message, 10);
	strcat(message, " POINTS");
	lcd_goto_xy(10, 0);
	lcd_str(message);
	lcd_goto_xy(10, 4);
	lcd_str("<ENTER>");
}

int main()
{	
	lcd_init();
	lcd_clear();
	lcd_update();
	lcd_contrast(0x38);
	
	init_btns();
	init_table();
	
	init_timer();
	
	while(1) {
		lcd_clear();
		/* Game logic */
		//btn_pressed();
		check_allignment();
		
		// 15 seconds have passed
		if (game_over >= ROUNDS) {
			display_game_over();
		} else {
			//display_grid();
			display_table();
		}
		
		lcd_delay();
		lcd_update();
	}
 
	return 0;
}